#include "ObjectManager.h"

#include <KDebug.h>

#include "Collisions.h"
#include "Debug.h"
#include "generated\DebugSettings.h"
#include "FunctionStubs.h"
#include "HUDManager.h"
#include "LevelManager.h"
#include "ScrollManager.h"
#include "MathUtils.h"
#include "exported\ExportedGGAnimations.h"
#include "PlayerManager.h"
#include "vdp_spr.h"
#include "VDP_Extra.h"
#include "DrawManager.h"


#define MAX_OBJECTS 90

ObjectStruct    objects[MAX_OBJECTS];
ObjectStruct*   objectVacancies[MAX_OBJECTS];
u16             numActiveObjects = 0;

ObjectStruct* firstObjectOfType[MAX_OBJECT_TYPES];
ObjectStruct* lastObjectOfType[MAX_OBJECT_TYPES];

ObjectStruct* firstNewObjectOfType[MAX_OBJECT_TYPES];
ObjectStruct* lastNewObjectOfType[MAX_OBJECT_TYPES];

void ObjectManager_Init()
{
    numActiveObjects = 0;

    u16 loop = 0;
    for (;loop < MAX_OBJECTS; loop++)
    {
        objectVacancies[loop] = &objects[loop];
    }

    memset(firstObjectOfType, NULL, sizeof(firstObjectOfType));
    memset(lastObjectOfType, NULL, sizeof(lastObjectOfType));
    memset(firstNewObjectOfType, NULL, sizeof(firstNewObjectOfType));
    memset(lastNewObjectOfType, NULL, sizeof(lastNewObjectOfType));
}

void PrintObjectList()
{
    Print("******** OBJECT LIST **********");
    u16 loop;

    for (loop = 0; loop < MAX_OBJECT_TYPES; loop++)
    {
        ObjectStruct* object = *(firstObjectOfType + loop);

        Print("Type");
        PrintNumber(loop);

        if (object != NULL)
        {
            while (object)
            {
                Print("Object");
                PrintNumber((u32)object);
                PrintNumber((u32)object->nextObjectOfSameType);
                PrintNumber((u32)object->previousObjectOfSameType);
                object = object->nextObjectOfSameType;
            }
        }

        Print("first object");
        PrintNumber((u32)*(firstObjectOfType + loop));
        Print("last object");
        PrintNumber((u32)*(lastObjectOfType + loop));
    }
    Print("******** END OBJECT LIST **********");
}

void PrintNewObjectList()
{
    Print("******** NEW OBJECT LIST **********");
    u16 loop;

    for (loop = 0; loop < MAX_OBJECT_TYPES; loop++)
    {
        ObjectStruct* object = *(firstNewObjectOfType + loop);

        Print("Type");
        PrintNumber(loop);

        if (object != NULL)
        {
            while (object)
            {
                Print("Object");
                PrintNumber((u32)object);
                PrintNumber((u32)object->nextObjectOfSameType);
                PrintNumber((u32)object->previousObjectOfSameType);

                object = object->nextObjectOfSameType;
            }
        }

        Print("first object");
        PrintNumber((u32)*(firstNewObjectOfType + loop));
        Print("last object");
        PrintNumber((u32)*(lastNewObjectOfType + loop));
    }
    Print("******** END NEW OBJECT LIST **********");
}

ObjectStruct* PrepareNewObject(u16 objectType)
{
    if (numActiveObjects >= MAX_OBJECTS)
    {
        return NULL;
    }

    // get the first available object slot 
    ObjectStruct* newObject = *(objectVacancies + numActiveObjects);

    // the next object of a new object is always NULL.
    newObject->nextObjectOfSameType = NULL;

    // if this is the first object of it's type, 
    // set the first new object to this.
    if (*(firstNewObjectOfType + objectType) == NULL)
    {
        *(firstNewObjectOfType + objectType) = newObject;
    }

    // if there's already a last object, attach ourselves to the last object 
    ObjectStruct* lastObject = *(lastNewObjectOfType + objectType);

    if (lastObject != NULL)
    {
        newObject->previousObjectOfSameType = lastObject;
        lastObject->nextObjectOfSameType = newObject;
    }

    // set ourselves as the last object.
    *(lastNewObjectOfType + objectType) = newObject;

    numActiveObjects++;

    newObject->objectType = objectType;

    return newObject;
}

ObjectStruct* ObjectManager_CreateObject(u16                            objectType, 
                                         ObjectFunctionType             updateFunction, 
                                         ObjectFunctionType             drawFunction,
                                         ObjectCollisionFunctionType    collisionFunction,
                                         ObjectCollisionFunctionType    grabbedFunction)
{
    ObjectStruct* newObject = PrepareNewObject(objectType);

    if (!newObject)
    {
        return NULL;
    }

    newObject->lifetime             = 0;
    newObject->updateFunction       = updateFunction;
    newObject->drawFunction         = drawFunction;
    newObject->collisionFunction    = collisionFunction;
    newObject->grabbedFunction      = grabbedFunction;


    /*
    Print("***********************");
    Print("Create Object");
    PrintObject(newObject);
    Print("first new object");
    PrintNumber((u32)*(firstNewObjectOfType + objectType));
    Print("last new object");
    PrintNumber((u32)*(lastNewObjectOfType + objectType));
    Print("***********************");
    */

    return newObject;
} 

void RemoveObject(ObjectStruct* object)
{
    /*
    Print("***********************");
    Print("Destroy Object");
    PrintObject(object);
    Print("first object");
    PrintNumber((u32)*(firstObjectOfType + object->objectType));
    Print("last object");
    PrintNumber((u32)*(lastObjectOfType + object->objectType));
    Print("***********************");
    */

    // if we have someone pointing at us, then make it point to
    // our next one.
    if (object->previousObjectOfSameType)
    {
        //Print("Connecting previous object to next object");
        object->previousObjectOfSameType->nextObjectOfSameType = object->nextObjectOfSameType;

        if (object->nextObjectOfSameType)
        {
            //Print("Connecting next object to previous object");
            object->nextObjectOfSameType->previousObjectOfSameType = object->previousObjectOfSameType;
        }
    }
    else
    // if we didn't have someone pointing at us, that means
    // we were first. change the first object of type to our next.
    {
        //Print("We were first. Changing first object to our next");

        *(firstObjectOfType + object->objectType) = object->nextObjectOfSameType;

        if (object->nextObjectOfSameType)
        {
            //Print("Setting next object to point to null previous object");
            object->nextObjectOfSameType->previousObjectOfSameType = NULL;
        }
    }

    // if we're pointing at an object, tell it to point at our previous.
    if (object->nextObjectOfSameType)
    {
        //Print("Changing next object to point to our previous");

        object->nextObjectOfSameType->previousObjectOfSameType = object->previousObjectOfSameType;

        if (object->previousObjectOfSameType)
        {
            //Print("Pointing previous's next to our next");
            object->previousObjectOfSameType->nextObjectOfSameType = object->nextObjectOfSameType;
        }
    }
    else
    // if we didn't have a next, that means we were last. move the last to the
    // object was was pointing to us.
    {
        //Print("We were last. Move the last object to our previous");

        *(lastObjectOfType + object->objectType) = object->previousObjectOfSameType;

        if (object->previousObjectOfSameType)
        {
            //Print("Setting our previous (new last) to point to null for its next");
            object->previousObjectOfSameType->nextObjectOfSameType = NULL;
        }
    }

    numActiveObjects--;
}

void ObjectManager_DestroyObject(ObjectStruct* object)
{
    RemoveObject(object);

    *(objectVacancies + numActiveObjects) = object;

    /*
    Print("***********************");
    Print("After Destroy Object");
    PrintObject(object);
    Print("first object");
    PrintNumber((u32)*(firstObjectOfType + object->objectType));
    Print("last object");
    PrintNumber((u32)*(lastObjectOfType + object->objectType));
    Print("***********************");
    */
}


void ObjectManager_ClearPendingObjects()
{
    memset(firstNewObjectOfType, NULL, sizeof(firstNewObjectOfType));
    memset(lastNewObjectOfType, NULL, sizeof(lastNewObjectOfType));
}

void ObjectManager_EraseAllObjectsOfType(u16 type)
{
    while (*(firstObjectOfType + type))
    {
        ObjectManager_DestroyObject(*(firstObjectOfType + type));
    }
}

void ProcessCollisions(u16 firstType, u16 secondType)
{
    if (*(firstObjectOfType + firstType )== NULL) return;
    if (*(firstObjectOfType + secondType) == NULL) return;

    ObjectStruct* firstObject = *(firstObjectOfType + firstType);
    ObjectStruct* secondObject = NULL;

    while (firstObject != NULL)
    {
        secondObject = *(firstObjectOfType + secondType);

        s16 left = firstObject->x + firstObject->rectLeft;
        s16 right = firstObject->x + firstObject->rectRight;
        s16 top = firstObject->y + firstObject->rectTop;
        s16 bottom = firstObject->y + firstObject->rectBottom;

        while (secondObject != NULL)
        {
            if (!(left > secondObject->x + secondObject->rectRight ||
                  right  < secondObject->x + secondObject->rectLeft ||
                  top    > secondObject->y + secondObject->rectBottom ||
                  bottom < secondObject->y + secondObject->rectTop))
            {
                secondObject->collisionFunction(secondObject, firstObject);
                firstObject->collisionFunction(firstObject, secondObject);

                if (!firstObject->healthPoints)
                {
                    goto nextprojectile;
                }
            }

            secondObject = secondObject->nextObjectOfSameType;
        }

nextprojectile:

        firstObject = firstObject->nextObjectOfSameType;
    }
}


void ProcessGrabber()
{
    if (grabbedObject)
    {
        return;
    }

    if (*(firstObjectOfType + TYPE_GRABBER)== NULL) return;
    if (*(firstObjectOfType + TYPE_ENEMY) == NULL) return;

    ObjectStruct* playerGrabber = *(firstObjectOfType + TYPE_GRABBER);
    ObjectStruct* object = NULL;

    while (playerGrabber != NULL)
    {
        object = *(firstObjectOfType + TYPE_ENEMY);

        s16 left = playerGrabber->x + playerGrabber->rectLeft;
        s16 right = playerGrabber->x + playerGrabber->rectRight;
        s16 top = playerGrabber->y + playerGrabber->rectTop;
        s16 bottom = playerGrabber->y + playerGrabber->rectBottom;

        while (object != NULL)
        {
            if (object->grabbedFunction)
            {
                if (!(left > object->x + object->rectRight ||
                    right  < object->x + object->rectLeft ||
                    top    > object->y + object->rectBottom ||
                    bottom < object->y + object->rectTop))
                {
                    object->grabbedFunction(object, playerGrabber);
                    playerGrabber->grabbedFunction(playerGrabber, object);

                    goto nextgrabber;
                }
            }

            object = object->nextObjectOfSameType;
        }

nextgrabber:

        playerGrabber = playerGrabber->nextObjectOfSameType;
    }
}


void ProcessCollisionsAgainstPlayer(u16 objectType)
{
    ObjectStruct* firstObject = *(firstObjectOfType + objectType);

    if (!firstObject) return;

    s16 playerRight = playerObject->x + playerObject->rectRight;
    s16 playerLeft= playerObject->x + playerObject->rectLeft;
    s16 playerBottom = playerObject->y + playerObject->rectBottom;
    s16 playerTop = playerObject->y + playerObject->rectTop;

    while (firstObject != NULL)
    {
        if (firstObject->x + firstObject->rectLeft   < playerRight &&
            firstObject->x + firstObject->rectRight  > playerLeft &&
            firstObject->y + firstObject->rectTop    < playerBottom &&
            firstObject->y + firstObject->rectBottom > playerTop)
        {
            firstObject->collisionFunction(firstObject, playerObject);
            PlayerManager_CollideWithPlayer(firstObject);
            return;
        }

        firstObject = firstObject->nextObjectOfSameType;
    }
}


void ProcessAllCollisions()
{
    ProcessCollisions(TYPE_PLAYER_BULLET, TYPE_ENEMY);
    ProcessCollisions(TYPE_PLAYER_PROJECTILE, TYPE_ENEMY);
    ProcessCollisions(TYPE_PLAYER_PROJECTILE, TYPE_ENEMY_PROJECTILE);
    ProcessGrabber();

    ProcessCollisionsAgainstPlayer(TYPE_PICKUP);

    if (debugSettings.collisions == 1)
    {
        // no need to do anything if we're dead.
        if (playerData.invincibilityTimer > 0 ||
            !playerObject->healthPoints)
        {
            return;
        }

        ProcessCollisionsAgainstPlayer(TYPE_ENEMY);
        ProcessCollisionsAgainstPlayer(TYPE_ENEMY_PROJECTILE);
    }
}

void AddNewObjects()
{
    // add new objects
    u16 loop = MAX_OBJECT_TYPES;

    while (loop--)
    {
        ObjectStruct* newObject = *(firstNewObjectOfType + loop);

        if (newObject != NULL)
        {

            // attach the last object's next to the new object
            // attach the new object's previous to the last object
            // move the last object to the last new object

            ObjectStruct* lastObject = *(lastObjectOfType + loop);
            if (lastObject != NULL)
            {
                lastObject->nextObjectOfSameType = newObject;
            }

            newObject->previousObjectOfSameType = lastObject;

            ObjectStruct* lastNewObject = *(lastNewObjectOfType + loop);

            *(lastObjectOfType + loop) = lastNewObject;

            if (*(firstObjectOfType + loop) == NULL)
            {
                *(firstObjectOfType + loop) = *(firstNewObjectOfType + loop);
            }
        }
    }

    memset(firstNewObjectOfType, NULL, sizeof(firstNewObjectOfType));
    memset(lastNewObjectOfType, NULL, sizeof(lastNewObjectOfType));
}


/*
// Don't call this function on objects that are on the new object list!!!
void ObjectManager_SwitchType(ObjectStruct* object, u16 toType)
{
    RemoveObject(object);
    ObjectStruct* blah = PrepareNewObject(toType);

    if (object != blah)
    {
        PrintObjectStruct(object);
        PrintObjectStruct(blah);
    }
}
*/

void UpdateObjects()
{
    u16 loop = 0;
    
    loop = MAX_OBJECT_TYPES;

    //u16 debugCounter = 0;

    // process current objects
    while (loop--)
    {
        ObjectStruct* object = *(firstObjectOfType + loop);

        while (object)
        {
            object->updateFunction(object);

            // this is probably optional
            object->lifetime++; // loops after 65535. Maximum of 18.2 minutes.

            object = object->nextObjectOfSameType;

            /*
            debugCounter++;

            if (debugCounter > 20)
            {
                Print("***************");
                PrintObjectStruct(object);
                PrintObjectList();
                PrintNewObjectList();

                while (1) {}
            }
            */
        }
    }
}

void ObjectManager_UpdateObjects()
{
        //Print("------------ Frame Begin -----------------");
    PlayerManager_UpdatePlayer();

        //Print("------------ Adding New Objects-----------------");
    AddNewObjects();

    //Print("------------ Updating Objects -----------------");
    UpdateObjects();

        //Print("------------ Object Collision6s -----------------");
    ProcessAllCollisions();

        //Print("------------ Frame End -----------------");
}

