#include "PlayerManager.h"

#include <genesis.h>
#include <kdebug.h>
#include <maths.h>
#include <vdp_spr.h>

#include "generated\AllObjects.h"

#include "GGAnimationDraw.h"
#include "Collisions.h"
#include "Config.h"
#include "Debug.h"
#include "generated\DebugSettings.h"
#include "EnemyTypes.h"
#include "exported\ExportedGenResSprites.h"
#include "exported\ExportedGGAnimations.h"
#include "HUDManager.h"
#include "JoypadManager.h"
#include "LevelManager.h"
#include "MathUtils.h"
#include "ObjectManager.h"
#include "objects\ObjectMovement.h"
#include "ScrollManager.h"
#include "StateManager.h"
#include "TileHeightMaps.h"
#include "VDP_Extra.h"
#include "WeaponManager.h"

typedef void (*VoidFunctionType)();

#define INIT_SCROLL_POSITION  0 // (302 * 4 * 8)

#define DEATH_FALL_TIME 120

// game constants.
#define STARTING_LIVES                  3
#define MAX_HEALTH_POINTS               10000
#define JUMP_SPEED                      25
#define KNOCKBACK_SPEED                 20
#define KNOCKBACK_UP_SPEED              -16
#define DEATH_JUMP_SPEED                17
#define MOVEMENT_SPEED                  6
#define DAMAGE_TIME                     10
#define INVINSIBILITY_TIME              120

#define GRAB_WAIT_TIME                  20

#define SPRITE_WAND_OFFSET_X (14)
#define SPRITE_WAND_OFFSET_Y (3)
#define SPRITE_WING_OFFSET_X (- 13)
#define SPRITE_WING_OFFSET_Y (3)

#define SCROLL_ACTIVATED

#ifdef SCROLL_ACTIVATED
    #define INITIAL_BACKGROUND_SCROLL_SPEED         2
#else
    #define INITIAL_BACKGROUND_SCROLL_SPEED         0
#endif

// helper macros
#define SET_ANIMATION(animation) \
    body.currentVDPInfo = &Player##animation##VDPInfo;\
    body.currentAnimation = &Player##animation; \
    body.animationFrame = 0;\
    body.animationSpeedCounter = 0;

#define NUM_WING_ANIMATION_FRAMES 4

const u16 const wingAnimationFrames[] = 
{
    0,
    1,
    2,
    1
};


s16 backgroundScrollSpeed = 0;
s16 backgroundScroll = INIT_SCROLL_POSITION;
s16 oldBackgroundScroll = INIT_SCROLL_POSITION;

u16 penguinFrame = 0;
u16 wandFrame = 0;
u16 wingFrame = 0;
s16 flyingOffsetY = 0;

u16 animationFrame = 0;
u16 animationSpeedCounter = 0;
u16 animationSpeed = 0;

u16 deathFallTimer = 0;
u16 deathWaitTime = 0;

u16 grabTimer = 0;
ObjectStruct* grabbedObject = NULL;

ObjectStruct playerStruct;
ObjectStruct* playerObject;
PlayerDataStruct playerData;

VoidFunctionType updateState = NULL;
VoidFunctionType updatePhysics = NULL;


struct
{
    const VDPInfo* currentVDPInfo;
    const GGAnimation* currentAnimation;
    u16 animationFrame;
    u16 animationSpeed;
    u16 animationSpeedCounter;
} body;



void StartEntering();
void UpdateEntering();

void StartFlying();
void UpdateFlying();
void UpdateFlyingPhysics();
//void StartFlyingAnimation();

void StartDeath();
void UpdateDeath();

void PlayerManager_AfterDeath();

void Nothing() {}

void StartEntering()
{
    updateState = UpdateEntering;
    updatePhysics = Nothing;
    animationSpeed = PlayerWings.frames[0]->frameDelayTime;
    HUDManager_UpdateLives();
}

void GrabOrThrow()
{
    if (grabbedObject)
    {
        // throw object
        ThrowGrabbedObject();
    }
    else // show grab animation
    {


        if (!grabTimer)
        {
            WandGrabber();
            grabTimer = GRAB_WAIT_TIME;
        }

    }
}

void Fire()
{
    if (!grabbedObject)
    {
        WeaponManager_Fire();
    }
    else
    {
        ThrowGrabbedObject();
    }
}

void UpdateEntering()
{
    if (!playerData.invincibilityTimer)
    {
        StartFlying();
        return;
    }

    playerObject->x += MOVEMENT_SPEED << 1;

    if (playerObject->x - backgroundScroll > P2V(120))
    {
        playerObject->x = backgroundScroll + P2V(120);
    }

    if (joystate & BUTTON_LEFT ||
        joystate & BUTTON_RIGHT ||
        joystate & BUTTON_UP ||
        joystate & BUTTON_DOWN)
    {
        StartFlying();
    }

    if (joystickState.aPressed)
    {
        Fire();
    }

    if (joystickState.bPressed)
    {
        WeaponManager_NextWeapon();
    }

    if (joystickState.cPressed)
    {
        GrabOrThrow();
    }
}

void StartFlying()
{
    playerData.playerState = PLAYERSTATE_FLYING;

    //StartFlyingAnimation();

    // updates
    updateState = UpdateFlying;
    updatePhysics = UpdateFlyingPhysics;

    // misc
    playerObject->speedx = 0;
}

void UpdateFlying()
{
    animationSpeed = PlayerWings.frames[0]->frameDelayTime;

    penguinFrame = 0;

    if (joystate & BUTTON_LEFT)
    {
        playerObject->speedx = -MOVEMENT_SPEED;

        animationSpeed = PlayerWings.frames[0]->frameDelayTime << 1;

        penguinFrame = 2;
    }
    else if (joystate & BUTTON_RIGHT)
    {
        playerObject->speedx = MOVEMENT_SPEED;

        animationSpeed = PlayerWings.frames[0]->frameDelayTime >> 1;

        penguinFrame = 1;
    }
    else
    {
        playerObject->speedx = 0;
    }

    if (joystate & BUTTON_UP)
    {
        playerObject->speedy = -MOVEMENT_SPEED;

        animationSpeed = PlayerWings.frames[0]->frameDelayTime >> 1;
    }
    else if (joystate & BUTTON_DOWN)
    {
        playerObject->speedy = MOVEMENT_SPEED;

        animationSpeed = PlayerWings.frames[0]->frameDelayTime << 1;
    }
    else
    {
        playerObject->speedy = 0;
    }

    if (joystickState.aPressed)
    {
        Fire();
    }

    if (joystickState.bPressed)
    {
        WeaponManager_NextWeapon();
    }

    if (joystickState.cPressed)
    {
        GrabOrThrow();
        //StartDeath();
        //StarPickupCreate(playerObject->x, playerObject->y + 512);
    }
}

/*
void StartFlyingAnimation()
{
    //SET_ANIMATION(Flying);
}
*/

void StartDeath()
{
    EffectCreate(&PlayerHit, playerObject->x - P2V(2), playerObject->y - P2V(2), 0);
    // updates
    updateState = UpdateDeath;
    updatePhysics = Nothing;

    if (grabbedObject)
    {
        ObjectManager_DestroyObject(grabbedObject);
        ObjectManager_DestroyObject((ObjectStruct*)grabbedObject->extra1);
        grabbedObject = NULL;
    }
}

void UpdateDeath()
{
    playerObject->speedx = INITIAL_BACKGROUND_SCROLL_SPEED << 1;

    playerObject->speedy++;

    if (playerObject->speedy > 20)
    {
        playerObject->speedy = 20;
    }

    playerObject->y += playerObject->speedy;
    playerObject->x += playerObject->speedx;

    if (playerObject->y > P2V(SCREEN_HEIGHT))
    {
        playerObject->y = P2V(SCREEN_HEIGHT);
    }

    deathFallTimer++;

    if (deathFallTimer > DEATH_FALL_TIME)
    {
        PlayerManager_SetupLife();
    }
}

const s16 touchPointLeftOffset = -PIXEL_TO_VIRTUAL(3);
const s16 touchPointRightOffset = PIXEL_TO_VIRTUAL(3);

void UpdateScreenBoundsChecking()
{
    if (playerObject->x + playerObject->rectLeft < foregroundScrollData.scrollx_hires_clamped)
    {
        playerObject->x = foregroundScrollData.scrollx_hires_clamped - playerObject->rectLeft;
    }

    if (playerObject->y + playerObject->rectTop < 0)
    {
        playerObject->y = -playerObject->rectTop;
    }

    if (playerObject->x + playerObject->rectRight > foregroundScrollData.scrollx_hires_clamped + PIXEL_TO_VIRTUAL(SCREEN_WIDTH))
    {
        playerObject->x = foregroundScrollData.scrollx_hires_clamped + PIXEL_TO_VIRTUAL(SCREEN_WIDTH) - playerObject->rectRight;
    }

    if (playerObject->y + playerObject->rectBottom > PIXEL_TO_VIRTUAL(SCREEN_HEIGHT))
    {
        playerObject->y = PIXEL_TO_VIRTUAL(SCREEN_HEIGHT) - playerObject->rectBottom;
    }
}

s16 effectiveSpeedX;

void UpdateLeftRightPhysics()
{
    s16 playerY = V2P(playerObject->y + playerObject->rectTop);
    s16 endXPosition = V2P(playerObject->x + effectiveSpeedX);

    if (effectiveSpeedX > 0)
    {
        endXPosition += V2P(playerObject->rectRight);

        if ((IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 8) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 20) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            playerObject->x = P2V(endXPosition - 1) - playerObject->rectRight;
            effectiveSpeedX = 0;
        }
    }
    else if (effectiveSpeedX < 0)
    {
        endXPosition += V2P(playerObject->rectLeft);

        if ((IsTileSolid_Pixel(endXPosition, playerY) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 8) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 16) != 0) ||
            (IsTileSolid_Pixel(endXPosition, playerY + 20) != 0))
        {
            endXPosition >>= 3;
            endXPosition <<= 3;
            playerObject->x = P2V(endXPosition) + playerObject->rectLeft;
            effectiveSpeedX = 0;
        }
    }
}

void UpdateUpDownPhysics()
{
    s16 playerX = V2P(playerObject->x + playerObject->rectLeft);
    s16 endYPosition = V2P(playerObject->y + playerObject->speedy);

    if (playerObject->speedy < 0)
    {
        endYPosition += V2P(playerObject->rectTop);

        if ((IsTileSolid_Pixel(playerX, endYPosition) != 0) || 
            (IsTileSolid_Pixel(playerX + 8, endYPosition) != 0) || 
            (IsTileSolid_Pixel(playerX + 10, endYPosition) != 0))
        {
            endYPosition >>= 3;
            endYPosition <<= 3;
            endYPosition += 8;

            playerObject->y = P2V(endYPosition) - playerObject->rectTop;
            playerObject->speedy = 0;
        }
    }
    else if (playerObject->speedy > 0)
    {
        endYPosition += V2P(playerObject->rectBottom);

        if ((IsTileSolid_Pixel(playerX, endYPosition) != 0) || 
            (IsTileSolid_Pixel(playerX + 8, endYPosition) != 0) || 
            (IsTileSolid_Pixel(playerX + 10, endYPosition) != 0))
        {
            endYPosition >>= 3;
            endYPosition <<= 3;

            playerObject->y = P2V(endYPosition - V2P(playerObject->rectBottom) - 1);
            playerObject->speedy = 0;
        }

    }
}

void InGeometryCheck()
{
    if (playerData.startingLifeTimer > 0)
    {
        return;
    }

    s16 playerX = V2P(playerObject->x + playerObject->rectLeft);
    s16 playerY = V2P(playerObject->y + playerObject->rectTop);

    if ((IsTileSolid_Pixel(playerX, playerY) != 0) ||
        (IsTileSolid_Pixel(playerX, playerY + 8) != 0) ||
        (IsTileSolid_Pixel(playerX, playerY + 16) != 0) ||
        (IsTileSolid_Pixel(playerX, playerY + 20) != 0) ||
        (IsTileSolid_Pixel(playerX + 10, playerY) != 0) ||
        (IsTileSolid_Pixel(playerX + 10, playerY + 8) != 0) ||
        (IsTileSolid_Pixel(playerX + 10, playerY + 16) != 0) ||
        (IsTileSolid_Pixel(playerX + 10, playerY + 20) != 0) ||
        (IsTileSolid_Pixel(playerX + 8, playerY) != 0) || 
        (IsTileSolid_Pixel(playerX + 8, playerY + 20) != 0))
    {
        StartDeath();
    }
}

void UpdateFlyingPhysics()
{
    UpdateScreenBoundsChecking();

    effectiveSpeedX = playerObject->speedx + backgroundScrollSpeed;

    if (playerData.startingLifeTimer > 0)
    {
        playerObject->x += effectiveSpeedX;
        playerObject->y += playerObject->speedy;
        return;
    }

    UpdateUpDownPhysics();

    playerObject->y += playerObject->speedy;

    UpdateLeftRightPhysics();

    playerObject->x += effectiveSpeedX;

    if (playerData.invincibilityTimer == 0 &&
        playerData.startingLifeTimer == 0)
    {
        InGeometryCheck();
    }


    /*
    // if the player is pushed past the left edge of the screen, we've
    // been squished.
    if (playerObject->x + playerObject->rectLeft < foregroundScrollData.scrollx_hires_clamped)
    {
        StartDeath();
    }
    */
}

void UpdateAutomated()
{
}

//#define SHOW_STATE_CHANGES

#ifdef SHOW_STATE_CHANGES
static u16 oldState = -1;
#endif 

void PlayerManager_UpdatePlayer()
{
#ifdef SHOW_STATE_CHANGES
    if (oldState != playerData.playerState)
    {
        switch (playerData.playerState)
        {
        case PLAYERSTATE_FLYING:
            Print("PLAYERSTATE_FLYING");
            break;
        }

        oldState = playerData.playerState;
    }
#endif
    u16 disabled = playerData.disablePlayer;

    if (disabled == 0)
    {
        playerData.oldPositionX = playerObject->x;
        playerData.oldPositionY = playerObject->y;

        oldBackgroundScroll = backgroundScroll;

        backgroundScroll += backgroundScrollSpeed;

        if (backgroundScrollSpeed > 0)
        {
            if (backgroundScroll > foregroundScrollData.scrollx_hires_limit)
            {
                backgroundScroll = foregroundScrollData.scrollx_hires_limit;
                backgroundScrollSpeed = 0;
            }
        }
        else if (backgroundScrollSpeed < 0)
        {
            if (backgroundScroll < 0)
            {
                backgroundScroll = INIT_SCROLL_POSITION;
            }
        }

        foregroundScrollData.scrollx_hires = backgroundScroll;

        updateState();
        updatePhysics();

        if (playerData.invincibilityTimer > 0)
        {
            playerData.invincibilityTimer--;
        }

        if (grabTimer)
        {
            grabTimer--;
        }

        if (playerData.startingLifeTimer > 0)
        {
            playerData.startingLifeTimer--;
        }

        WeaponManager_Update();
    }

    playerObject->lifetime++;

    PlayerManager_DrawPlayer();
}

void PlayerManager_CollideWithPlayer(ObjectStruct* collidedObject)
{
    if (playerData.invincibilityTimer > 0)
    {
        return;
    }

    if (collidedObject->objectType == TYPE_ENEMY ||
        collidedObject->objectType == TYPE_ENEMY_PROJECTILE)
    {
        if (playerObject->healthPoints > 0)
        {
            if (debugSettings.invincible == 0)
            {

                if (numStars[currentWeapon])
                {
                    StarFallEffect(playerObject->x, playerObject->y, -3, -24);
                    StarFallEffect(playerObject->x, playerObject->y, 3, -24);
                    StarFallEffect(playerObject->x, playerObject->y, -2, -25);
                    StarFallEffect(playerObject->x, playerObject->y, 2, -25);
                    StarFallEffect(playerObject->x, playerObject->y, 0, -26);
                }
            
                WeaponManager_RemoveStars(25);

                playerObject->healthPoints--;
            }

            EffectCreate(&PlayerHit, playerObject->x - P2V(2), playerObject->y - P2V(2), 0);

            if (playerObject->healthPoints == 0)
            {
                StartDeath();
            }
            else
            {
                playerData.invincibilityTimer = INVINSIBILITY_TIME;
            }
        }
    }
}

void UpdateAnimation()
{
    animationSpeedCounter++;
    if (animationSpeedCounter > animationSpeed)
    {
        animationFrame++;
        animationSpeedCounter = 0;
    }

    if (animationFrame > NUM_WING_ANIMATION_FRAMES - 1)
    {
        animationFrame = 0;
    }

    if (playerData.invincibilityTimer & 1)
    {
        return;
    }

    flyingOffsetY = -wingAnimationFrames[animationFrame] + 1;

    if (!grabTimer && !grabbedObject)
    {
        wandFrame = (weaponFireCooldownTimer > 0);

        if (weaponFireCooldownTimer > FIRE_COOLDOWN_TIME - 3)
        {
            wandFrame = 2;
        }
    }
    else
    {
        wandFrame = 1;
    }

    wingFrame = wingAnimationFrames[animationFrame];


    GGAnimation_Draw(playerObject->screenx, 
                     playerObject->screeny + flyingOffsetY, 
                     &Player, 
                     penguinFrame, 
                     playerObject->tileAttribute);

    GGAnimation_Draw(playerObject->screenx + SPRITE_WAND_OFFSET_X, 
                     playerObject->screeny + SPRITE_WAND_OFFSET_Y + flyingOffsetY, 
                     &PlayerWand, 
                     wandFrame, 
                     TILE_ATTR(PAL1, 1, 0, 0));

    GGAnimation_Draw(playerObject->screenx + SPRITE_WING_OFFSET_X, 
                     playerObject->screeny + SPRITE_WING_OFFSET_Y + flyingOffsetY, 
                     &PlayerWings, 
                     wingAnimationFrames[animationFrame], 
                     playerObject->tileAttribute);
}

void PlayerManager_DrawPlayer()
{
    // update screen position
    playerObject->screenx = (s16)VIRTUAL_TO_PIXEL(playerObject->x - foregroundScrollData.scrollx_hires_clamped);
    playerObject->screeny = (s16)VIRTUAL_TO_PIXEL(playerObject->y - foregroundScrollData.scrolly_hires_clamped);

    //playerObject->screenx = (s16)VIRTUAL_TO_PIXEL(playerObject->x);
    //playerObject->screeny = (s16)VIRTUAL_TO_PIXEL(playerObject->y);


    /*
    // debugging
    if (debugSettings.showCollisionBox == 1)
    {
        MyVDP_setSprite(playerObject->screenx - 4, playerObject->screeny + 24, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->screenx - 4, playerObject->screeny + 16, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->screenx - 4, playerObject->screeny + 8, SPRITE_SIZE(1,1), 1);
        MyVDP_setSprite(playerObject->screenx - 4, playerObject->screeny, SPRITE_SIZE(1,1), 1);
    }
    */

    UpdateAnimation();
}

void PlayerManager_InitPosition(s16 x, s16 y)
{
    playerData.playerLevelStartX = x;
    playerData.playerLevelStartY = y;
    playerData.oldPositionX = x;
    playerData.oldPositionY = y;
    oldBackgroundScroll = INIT_SCROLL_POSITION;
    backgroundScroll = INIT_SCROLL_POSITION;
}

void PlayerManager_SetupForGame()
{
    playerObject = &playerStruct;

    playerObject->objectType        = TYPE_PLAYER;
    playerObject->damage            = 1;
    playerObject->tileAttribute     = TILE_ATTR(PAL0, 1, 0, 0);

    playerObject->rectLeft          = PIXEL_TO_VIRTUAL(4);
    playerObject->rectTop           = PIXEL_TO_VIRTUAL(2);
    playerObject->rectRight         = PIXEL_TO_VIRTUAL(14);
    playerObject->rectBottom        = PIXEL_TO_VIRTUAL(22);
     
    playerData.lives                = STARTING_LIVES;
    playerData.playerLevelStartX    = 120;
    playerData.playerLevelStartY    = 120;
    playerData.oldPositionX         = 0;
    playerData.oldPositionY         = 0;

    oldBackgroundScroll = INIT_SCROLL_POSITION;
    backgroundScroll = INIT_SCROLL_POSITION;

    backgroundScrollSpeed = INITIAL_BACKGROUND_SCROLL_SPEED;

    penguinFrame = 0;
    wandFrame = 0;
    wingFrame = 0;
    flyingOffsetY = 0;

    animationFrame = 0;
    animationSpeedCounter = 0;
    animationSpeed = 0;

    deathFallTimer = 0;
    deathWaitTime = 0;
}

void PlayerManager_AfterDeath()
{
    playerObject->healthPoints      = MAX_HEALTH_POINTS;
}

void PlayerManager_SetupLife()
{
    playerObject->x                 = backgroundScroll;
    playerObject->y                 = PIXEL_TO_VIRTUAL(70);
    playerObject->speedx            = 0;
    playerObject->speedy            = 0;
    playerObject->healthPoints      = MAX_HEALTH_POINTS;

    playerData.invincibilityTimer   = INVINSIBILITY_TIME;
    playerData.startingLifeTimer    = INVINSIBILITY_TIME;
    playerData.freezeTimer          = 0;
    playerData.onfreezeTimerDone    = NULL;
    playerData.playerState          = 0;
    playerData.disablePlayer        = 0;
    playerData.platform             = NULL;
    playerData.oldPositionX         = playerObject->x;
    playerData.oldPositionY         = playerObject->y;

    deathWaitTime = 0;
    deathFallTimer = 0;

    grabbedObject = NULL;

    WeaponManager_Init();

    StartEntering();
}

